// Copyright 2025 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

#include "textflag.h"

DATA	·x(SB)/8, $123
GLOBL	·x(SB), NOPTR, $8
