// Copyright 2024 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package main

/*
void f(){}
int g(double x){}
*/
import "C"

func init() {
	C.f()
	C.g(0)
}
