// Code generated by mkmalloc.go; DO NOT EDIT.
//go:build !plan9

package runtime

import "unsafe"

var mallocScanTable = [513]func(size uintptr, typ *_type, needzero bool) unsafe.Pointer{
	mallocPanic,
	mallocgcSmallScanNoHeaderSC1,
	mallocgcSmallScanNoHeaderSC1,
	mallocgcSmallScanNoHeaderSC1,
	mallocgcSmallScanNoHeaderSC1,
	mallocgcSmallScanNoHeaderSC1,
	mallocgcSmallScanNoHeaderSC1,
	mallocgcSmallScanNoHeaderSC1,
	mallocgcSmallScanNoHeaderSC1,
	mallocgcSmallScanNoHeaderSC2,
	mallocgcSmallScanNoHeaderSC2,
	mallocgcSmallScanNoHeaderSC2,
	mallocgcSmallScanNoHeaderSC2,
	mallocgcSmallScanNoHeaderSC2,
	mallocgcSmallScanNoHeaderSC2,
	mallocgcSmallScanNoHeaderSC2,
	mallocgcSmallScanNoHeaderSC2,
	mallocgcSmallScanNoHeaderSC3,
	mallocgcSmallScanNoHeaderSC3,
	mallocgcSmallScanNoHeaderSC3,
	mallocgcSmallScanNoHeaderSC3,
	mallocgcSmallScanNoHeaderSC3,
	mallocgcSmallScanNoHeaderSC3,
	mallocgcSmallScanNoHeaderSC3,
	mallocgcSmallScanNoHeaderSC3,
	mallocgcSmallScanNoHeaderSC4,
	mallocgcSmallScanNoHeaderSC4,
	mallocgcSmallScanNoHeaderSC4,
	mallocgcSmallScanNoHeaderSC4,
	mallocgcSmallScanNoHeaderSC4,
	mallocgcSmallScanNoHeaderSC4,
	mallocgcSmallScanNoHeaderSC4,
	mallocgcSmallScanNoHeaderSC4,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC5,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC6,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC7,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC8,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC9,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC10,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC11,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC12,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC13,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC14,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC15,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC16,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC17,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC18,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC19,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC20,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC21,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC22,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC23,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC24,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC25,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
	mallocgcSmallScanNoHeaderSC26,
}

var mallocNoScanTable = [513]func(size uintptr, typ *_type, needzero bool) unsafe.Pointer{
	mallocPanic,
	mallocgcTinySize1,
	mallocgcTinySize2,
	mallocgcTinySize3,
	mallocgcTinySize4,
	mallocgcTinySize5,
	mallocgcTinySize6,
	mallocgcTinySize7,
	mallocgcTinySize8,
	mallocgcTinySize9,
	mallocgcTinySize10,
	mallocgcTinySize11,
	mallocgcTinySize12,
	mallocgcTinySize13,
	mallocgcTinySize14,
	mallocgcTinySize15,
	mallocgcSmallNoScanSC2,
	mallocgcSmallNoScanSC3,
	mallocgcSmallNoScanSC3,
	mallocgcSmallNoScanSC3,
	mallocgcSmallNoScanSC3,
	mallocgcSmallNoScanSC3,
	mallocgcSmallNoScanSC3,
	mallocgcSmallNoScanSC3,
	mallocgcSmallNoScanSC3,
	mallocgcSmallNoScanSC4,
	mallocgcSmallNoScanSC4,
	mallocgcSmallNoScanSC4,
	mallocgcSmallNoScanSC4,
	mallocgcSmallNoScanSC4,
	mallocgcSmallNoScanSC4,
	mallocgcSmallNoScanSC4,
	mallocgcSmallNoScanSC4,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC5,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC6,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC7,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC8,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC9,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC10,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC11,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC12,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC13,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC14,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC15,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC16,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC17,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC18,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC19,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC20,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC21,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC22,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC23,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC24,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC25,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
	mallocgcSmallNoScanSC26,
}
