// Code generated by mkpreempt.go; DO NOT EDIT.

package runtime

type xRegs struct {
	X0  [32]byte
	X1  [32]byte
	X2  [32]byte
	X3  [32]byte
	X4  [32]byte
	X5  [32]byte
	X6  [32]byte
	X7  [32]byte
	X8  [32]byte
	X9  [32]byte
	X10 [32]byte
	X11 [32]byte
	X12 [32]byte
	X13 [32]byte
	X14 [32]byte
	X15 [32]byte
	X16 [32]byte
	X17 [32]byte
	X18 [32]byte
	X19 [32]byte
	X20 [32]byte
	X21 [32]byte
	X22 [32]byte
	X23 [32]byte
	X24 [32]byte
	X25 [32]byte
	X26 [32]byte
	X27 [32]byte
	X28 [32]byte
	X29 [32]byte
	X30 [32]byte
	X31 [32]byte
}
