// Copyright 2025 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build goexperiment.simd

package simd

// Invoke code generators.
// The test helpers are generated by tmplgen from the archsimd package.

//go:generate go -C ../.. generate
