// Copyright 2022 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package p

func f[P *Q, Q any](p P, q Q) {
	func() {
		_ = f[P]
		f(p, q)
		f[P](p, q)
		f[P, Q](p, q)
	}()
}
